#ifdef WIN32
# include <windows.h>
#endif

#include <GL/gl.h>
#undef GL_GLEXT_PROTOTYPES


#define GLH_EXT_SINGLE_FILE
#define GLH_NVEB_USING_NVPARSE

#include <glh_nveb.h>
#include <glh_extensions.h>
#include <glh_obs.h>
#include <glh_glut.h>
#include <nvparse.h>
#include <math.h>

NVEB_EFFECT_NAME("Depth Sprites");
NVEB_EFFECT_VERSION("1.0");
NVEB_EFFECT_LOCATION("Effects\\Imposters");
NVEB_EFFECT_ABOUT(0, NULL, "Depth Sprites", "NVEffectsExplained.htm#DepthSprites");
NVEB_EFFECT_ABOUT(1, "Date", "May 2001", NULL);
NVEB_EFFECT_ABOUT(2, NULL, "NVIDIA Developer Relations", "http://www.nvidia.com/Developer");

using namespace glh;

// GLH controls
glut_callbacks cb;
glut_simple_mouse_interactor camera, object;
glut_perspective_reshaper reshaper;

// Global variables
int dim = 4;                  // Number of sprites that are drawn on a side - change using 'D' or 'd'
float light_radius = 1.5;     // Determines orbit height (or radius) - change using 'L' or 'l'
float theta = 0.;             // Used for animating the light as an orbit around the sprites
float sphere_radius = 1.5 / dim; // Radius of spheres/sprites used
bool b[256];                  // Toggle state for all the keys (characters)

int framecount = 0;
LARGE_INTEGER oldCount, newCount;
LARGE_INTEGER frequency;


GLuint vertex_program_id;      // Vertex program handle
GLuint bind_vertex_program; // Display list handle for initializing vertex program attributes
GLuint bind_texture_shaders_no_depth; // Display list handle for initializing trivial texture shader
GLuint bind_texture_shaders_depth;    // Display list handle for initializing real texture shader
GLuint bind_diffuse_register_combiners; // Display list handle for initializing register combiners (diffuse only)
GLuint bind_diffuse_with_normalization_register_combiners;
GLuint bind_specular_register_combiners; // Display list handle for initializing register combiners (diffuse + specular)
GLuint bind_specular_with_normalization_register_combiners;
GLuint draw_solid_sphere;     // Display list handle for drawing a single sphere at origin

tex_object_2D normals_ubyte;
tex_object_2D depth_hilo;

// Constants
const float PI = 3.141592654;
const float theta_step = .1;
const float light_radius_step = .1;
const float min_light_radius = .5;

// glut-ish callbacks
void display();
void key(unsigned char k, int x, int y);
void menu( int v );
void idle();

// my functions
void init_opengl();

















#define  A		0
#define  B		1
#define  C		2
#define  D		3

#define  rand3a(x,y,z)	frand(67*(x)+59*(y)+71*(z))
#define  rand3b(x,y,z)	frand(73*(x)+79*(y)+83*(z))
#define  rand3c(x,y,z)	frand(89*(x)+97*(y)+101*(z))
#define  rand3d(x,y,z)	frand(103*(x)+107*(y)+109*(z))

/* hermite */
#define  hpoly1(t)	((2.0*t-3.0)*t*t+1.0)
#define  hpoly2(t)	(-2.0*t+3.0)*t*t
#define  hpoly3(t)	((t-2.0)*t+1.0)*t
#define  hpoly4(t)	(t-1.0)*t*t

double  *noise3(), fnoise3(), frand();

static  interpolate();

static long  xlim[3][2];
static double  xarg[3];

#define  EPSILON	.0001		/* error allowed in fractal */

//#define  frand3(x,y,z)	frand(17*(x)+23*(y)+29*(z))
#define  frand3(x,y,z)	frand(37*(x)+132*(y)+291*(z))


double frand( register long s )			/* get random number from seed */
{
	s = s<<13 ^ s;
	return(1.0-((s*(s*s*15731+789221)+1376312589)&0x7fffffff)/1073741824.0);
}


static interpolate( double f[4], register int i, register int n)
{
	double  f0[4], f1[4], hp1, hp2;

	if (n == 0) {
		f[A] = rand3a(xlim[0][i&1],xlim[1][i>>1&1],xlim[2][i>>2]);
		f[B] = rand3b(xlim[0][i&1],xlim[1][i>>1&1],xlim[2][i>>2]);
		f[C] = rand3c(xlim[0][i&1],xlim[1][i>>1&1],xlim[2][i>>2]);
		f[D] = rand3d(xlim[0][i&1],xlim[1][i>>1&1],xlim[2][i>>2]);
	} else {
		n--;
		interpolate(f0, i, n);
		interpolate(f1, i | 1<<n, n);
		hp1 = hpoly1(xarg[n]); hp2 = hpoly2(xarg[n]);
		f[A] = f0[A]*hp1 + f1[A]*hp2;
		f[B] = f0[B]*hp1 + f1[B]*hp2;
		f[C] = f0[C]*hp1 + f1[C]*hp2;
		f[D] = f0[D]*hp1 + f1[D]*hp2 +
				f0[n]*hpoly3(xarg[n]) + f1[n]*hpoly4(xarg[n]);
	}
}

double *noise3(register double xnew[3])			/* compute the noise function */
{
	static double  x[3] = {-100000.0, -100000.0, -100000.0};
	static double  f[4];

	if (x[0]==xnew[0] && x[1]==xnew[1] && x[2]==xnew[2])
		return(f);
	x[0] = xnew[0]; x[1] = xnew[1]; x[2] = xnew[2];
	xlim[0][0] = (long)floor(x[0]); xlim[0][1] = xlim[0][0] + 1;
	xlim[1][0] = (long)floor(x[1]); xlim[1][1] = xlim[1][0] + 1;
	xlim[2][0] = (long)floor(x[2]); xlim[2][1] = xlim[2][0] + 1;
	xarg[0] = x[0] - xlim[0][0];
	xarg[1] = x[1] - xlim[1][0];
	xarg[2] = x[2] - xlim[2][0];
	interpolate(f, 0, 3);
	return(f);
}





double fnoise3( vec3f pos )			/* compute fractal noise function */
{
	double  p[3];
	long  t[3], v[3], beg[3];
	double  fval[8], fc;
	int  branch;
	register long  s;
	register int  i, j;

	p[0]=pos[0];
	p[1]=pos[1];
	p[2]=pos[2];
	/* get starting cube */
	s = (long)(1.0/EPSILON);
	for (i = 0; i < 3; i++) {
		t[i] = (long)(s*p[i]);
		beg[i] = s*(long)(floor(p[i]));
	}
	for (j = 0; j < 8; j++) {
		for (i = 0; i < 3; i++) {
			v[i] = beg[i];
			if (j & 1<<i)
				v[i] += s;
		}
		fval[j] = frand3(v[0],v[1],v[2]);
	}
						/* compute fractal */
	for ( ; ; ) {
		fc = 0.0;
		for (j = 0; j < 8; j++)
			fc += fval[j];
		fc *= 0.125;
		if ((s >>= 1) == 0)
		{
			if(fc<-1.0)fc=-1.0;
			if(fc>1.0)fc=1.0;
			return(fc);		/* close enough */
		}
		branch = 0;
		for (i = 0; i < 3; i++) {	/* do center */
			v[i] = beg[i] + s;
			if (t[i] > v[i]) {
				branch |= 1<<i;
			}
		}
		fc += s*EPSILON*frand3(v[0],v[1],v[2]);
		fval[~branch & 7] = fc;
		for (i = 0; i < 3; i++) {	/* do faces */
			if (branch & 1<<i)
				v[i] += s;
			else
				v[i] -= s;
			fc = 0.0;
			for (j = 0; j < 8; j++)
				if (~(j^branch) & 1<<i)
					fc += fval[j];
			fc = 0.25*fc + s*EPSILON*frand3(v[0],v[1],v[2]);
			fval[~(branch^1<<i) & 7] = fc;
			v[i] = beg[i] + s;
		}
		for (i = 0; i < 3; i++) {	/* do edges */
			j = (i+1)%3;
			if (branch & 1<<j)
				v[j] += s;
			else
				v[j] -= s;
			j = (i+2)%3;
			if (branch & 1<<j)
				v[j] += s;
			else
				v[j] -= s;
			fc = fval[branch & ~(1<<i)];
			fc += fval[branch | 1<<i];
			fc = 0.5*fc + s*EPSILON*frand3(v[0],v[1],v[2]);
			fval[branch^1<<i] = fc;
			j = (i+1)%3;
			v[j] = beg[j] + s;
			j = (i+2)%3;
			v[j] = beg[j] + s;
		}
		for (i = 0; i < 3; i++)		/* new cube */
			if (branch & 1<<i)
				beg[i] += s;
	}
}
























int main(int argc, char **argv)
{
	QueryPerformanceFrequency(&frequency);


    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB );
    glutCreateWindow("Dot Product Depth Replace (Depth Sprite) Demo");

    b['w'] = false;
    b['a'] = true;
    b['z'] = true;
    b['t'] = true;
    b['s'] = true;
    b['g'] = false;
    b['n'] = true;
    b[' '] = true;

    init_opengl();

    glut_helpers_initialize();

    cb.keyboard_function = key;
    camera.configure_buttons(1);
    camera.set_camera_mode(true);
    object.configure_buttons(1);

    // Next two lines are to start up the demo in an interesting position
    // Please forgive the seemingly frivolous use of magic numbers
    object.dolly.dolly[2] = -3;
    object.trackball.r.set_value(vec3f(1, 0, 0), -PI/4.);

    glut_add_interactor(&cb);
    glut_add_interactor(&object);
    glut_add_interactor(&reshaper);

    glutCreateMenu( menu );
    glutAddMenuEntry( "toggle wireframe [w]", 'w' );
    glutAddMenuEntry( "toggle depth sprites [t]", 't' );
    glutAddMenuEntry( "toggle alpha test [a]", 'a' );
    glutAddMenuEntry( "toggle depth test [z]", 'z' );
    glutAddMenuEntry( "toggle light animation [ ]", ' ' );
    glutAddMenuEntry( "toggle specular lighting [s]", 's' );
    glutAddMenuEntry( "toggle per-pixel normalization [n]", 'n' );
    glutAddMenuEntry( "toggle drawing real geometry [g]", 'g' );
    glutAddMenuEntry( "increase number of depth sprites [d]", 'd' );
    glutAddMenuEntry( "decrease number of depth sprites [D]", 'D' );
    glutAddMenuEntry( "increase light's orbital radius [l]", 'l' );
    glutAddMenuEntry( "decrease light's orbital radius [L]", 'L' );
    glutAttachMenu( GLUT_RIGHT_BUTTON );

    glutIdleFunc(idle);
    glutDisplayFunc(display);
    glutMainLoop();
    return 0;
}

static void make_normals_texture()
{
    int imgsize = 256;
    GLubyte * uimg = new GLubyte[imgsize*imgsize*4];
    GLubyte * uip = uimg;

    for(int j=0; j < imgsize; j++)
    {
        float fj = 2*j/(imgsize-1.f) - 1;
        for(int i=0; i < imgsize; i++)
        {
            float fi = 2*i/(imgsize-1.f) - 1;
            vec3f n;
            GLubyte alpha = 255;
            n[0] = fi;
            n[1] = fj;
            n[2] = sqrt(1 - fi*fi - fj*fj);
            if(n[2] <= 0)
            {
                n = vec3f(0,0,1);
                alpha = 0;
            }
//            n[2] += ((sin(j*0.3f)*0.5f+cos(i*0.18f)*0.5f)*0.1+0.1f); // depth

			vec3f cpy=n;

			n[0]+=0.1f*fnoise3(cpy*3.0f);
			n[1]+=0.1f*fnoise3(cpy*3.0f+vec3f(10.423,22.543,21.534));
			n[2]+=0.1f*fnoise3(cpy*3.0f+vec3f(100.44342,2121.33,33.1));

			n[0]+=0.2f*fnoise3(cpy*13.0f);
			n[1]+=0.2f*fnoise3(cpy*13.0f+vec3f(10.423,22.543,21.534));
			n[2]+=0.2f*fnoise3(cpy*13.0f+vec3f(100.44342,2121.33,33.1));

            n.normalize();
          *uip++ = n[0] * 127 + 128;
          *uip++ = n[1] * 127 + 128;
          *uip++ = n[2] * 127 + 128;
//          *uip++ = n[0] * 117 + 128+(rand()%20)-10;
//          *uip++ = n[1] * 117 + 128+(rand()%20)-10;
//          *uip++ = n[2] * 117 + 128+(rand()%20)-10;
            *uip++ = alpha;

        }
    }

    normals_ubyte.bind();
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, imgsize, imgsize, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, uimg);
    normals_ubyte.parameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    delete [] uimg;
}

static GLfloat clamp_0_to_1(GLfloat x)
{
    return (fabs(x) - fabs(x-1.)) * .5 + .5;
}

static void make_depth_texture()
{
    int imgsize = 256;
    GLushort * img = new GLushort[imgsize*imgsize*2];
    GLushort * ip = img;

    for(int j=0; j < imgsize; j++)
    {
        float fj = 2*j/(imgsize-1.f) - 1;
        for(int i=0; i < imgsize; i++)
        {
            float fi = 2*i/(imgsize-1.f) - 1;
            vec3f n;
            n[0] = fi;
            n[1] = fj;
            n[2] = clamp_0_to_1(sqrt(1. - fi*fi - fj*fj));
            n.normalize();
          *ip++ = n[2] * 65535; // depth
//            *ip++ = n[2] * (65535-2000)+((sin(j*0.3f)*0.5f+cos(i*0.18f)*0.5f)*1000+1000); // depth
            *ip++ = 0; // not used
        }
    }

    depth_hilo.bind();

    glTexImage2D(GL_TEXTURE_2D, 0, GL_HILO_NV, imgsize, imgsize, 0,
                 GL_HILO_NV, GL_UNSIGNED_SHORT, img);
    depth_hilo.parameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    depth_hilo.parameter(GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    delete [] img;
}

static void create_sphere_dl()
{
       draw_solid_sphere = glGenLists(1);
    sphere_radius = 1.5 / (float)dim;
       glNewList(draw_solid_sphere, GL_COMPILE);
    glutSolidSphere(sphere_radius, 100, 100);
    glEndList();
}

static void init_opengl()
{
    glClearColor(.25f, .25f, .25f, 1);
    if (!glh_init_extensions("\
         GL_EXT_secondary_color \
         GL_ARB_multitexture \
         GL_NV_vertex_program \
         GL_NV_texture_shader \
         GL_NV_register_combiners \
        "))
    {
        cout << "Demo requires the following extension(s): " << glh_get_unsupported_extensions() << "\n";
        exit(0);
    }
    make_normals_texture();
    make_depth_texture();

    create_sphere_dl();

    glGenProgramsNV(1, &vertex_program_id);
    glBindProgramNV(GL_VERTEX_PROGRAM_NV, vertex_program_id);

    nvparse(
        "!!VP1.0\n"
        "# R2.x = radius;\n"
        "MOV R2.x, v[TEX0].w;\n"

        "# R3 = [center.x, center.y, center.z, 1];\n"
        "MOV R3.xyz, v[TEX0];\n"
        "MOV R3.w, c[16].y;\n"

        "# z_scale = radius * PROJ[10];\n"
        "MUL o[TEX1].x, R2.x, c[10].z;\n"

        "# z_bias = MVP[2] * center;\n"
        "DP4 o[TEX1].z, c[2], R3;\n"

        "# w_scale = radius * PROJ[11];\n"
        "MUL o[TEX2].x, R2.x, c[11].z;\n"

        "# w_bias = MVP[3] * center;\n"
        "DP4 o[TEX2].z, c[3], R3;\n"

        "# eye_space_offset = radius * v[OPOS];\n"
        "MUL R1, R2.x, v[OPOS];\n"

        "# object_space_offset = INV_MV * eye_space_offset;\n"
        "DP4 R4.x, c[4], R1;\n"
        "DP4 R4.y, c[5], R1;\n"
        "DP4 R4.z, c[6], R1;\n"

        "# center += object_space_offset ;\n"
        "ADD R3.xyz, R3, R4;\n"
      
        "# clip_coord = MVP * center;\n"
        "DP4 o[HPOS].x, c[0], R3;\n"
        "DP4 o[HPOS].y, c[1], R3;\n"
        "DP4 o[HPOS].z, c[2], R3;\n"
        "DP4 o[HPOS].w, c[3], R3;\n"

        "# texcoord[0] = texcoord[3] = v[OPOS] * .5 + .5;\n"
        "MAD o[TEX0].xy, v[OPOS], c[16].x, c[16].x;\n"
        "MAD o[TEX3].xy, v[OPOS], c[16].x, c[16].x;\n"

        "# move color from input to output\n"
        "MOV o[COL0], v[COL0];\n"

        "# get vector from point to light\n"
        "ADD R3.xyz, v[COL1], -R3;\n"

        "# transform light vector into eye space\n"
        "DP4 R5.x, c[12], R3;\n"
        "DP4 R5.y, c[13], R3;\n"
        "DP4 R5.z, c[14], R3;\n"

        "# normalize light vector \n"
        "DP3 R3, R5, R5;\n"
        "RSQ R3, R3.x;\n"
        "MUL R5, R5, R3;\n"

        "# range compress light vector to fit in a color \n"
        "MAD o[COL1].xyz, R5, c[16].x, c[16].x;\n"

        "END"
    );
    for (char * const * errors = nvparse_get_errors(); *errors; errors++)
        fprintf(stderr, *errors);

    bind_vertex_program = glGenLists(1);
    glNewList(bind_vertex_program, GL_COMPILE);
    glBindProgramNV(GL_VERTEX_PROGRAM_NV, vertex_program_id);
    glTrackMatrixNV(GL_VERTEX_PROGRAM_NV, 0, GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV);
    glTrackMatrixNV(GL_VERTEX_PROGRAM_NV, 4, GL_MODELVIEW, GL_INVERSE_NV);
    glTrackMatrixNV(GL_VERTEX_PROGRAM_NV, 8, GL_PROJECTION, GL_IDENTITY_NV);
    glTrackMatrixNV(GL_VERTEX_PROGRAM_NV, 12, GL_MODELVIEW, GL_INVERSE_TRANSPOSE_NV);
    glProgramParameter4fNV(GL_VERTEX_PROGRAM_NV, 16, .5, 1, 0, 0);
    glEnable(GL_VERTEX_PROGRAM_NV);
    glEndList();

    bind_texture_shaders_depth = glGenLists(1);
    glNewList(bind_texture_shaders_depth, GL_COMPILE);
    // bind textures
    // stage 0 -- unsigned HILO depth texture
    glActiveTextureARB( GL_TEXTURE0_ARB );
    depth_hilo.bind();
    // stage 3 -- normal map
    glActiveTextureARB( GL_TEXTURE3_ARB );
    normals_ubyte.bind();
    // Configure texture shaders
    nvparse(
        "!!TS1.0                                           \n"
        "// Depth sprite plus normal map for lighting      \n"
        "texture_2d(); // depth image, unsigned HILO 16/16 \n"
        "dot_product_depth_replace_1of2(tex0);             \n"
        "dot_product_depth_replace_2of2(tex0);             \n"
        "texture_2d(); // normal map                       \n"
    );
    for (errors = nvparse_get_errors(); *errors; errors++)
        fprintf(stderr, "Texture shader error: %s\n", *errors);
    glEnable(GL_TEXTURE_SHADER_NV);
    glEndList();

    bind_texture_shaders_no_depth = glGenLists(1);
    glNewList(bind_texture_shaders_no_depth, GL_COMPILE);
    // bind textures
    // stage 3 -- normal map
    glActiveTextureARB( GL_TEXTURE3_ARB );
    normals_ubyte.bind();
    // Configure texture shaders
    nvparse(
        "!!TS1.0                                           \n"
        "// No depth sprite, just normal map for lighting  \n"
        "nop();                                            \n"
        "nop();                                            \n"
        "nop();                                            \n"
        "texture_2d();                                     \n"
    );
    for (errors = nvparse_get_errors(); *errors; errors++)
        fprintf(stderr, "Texture shader error: %s\n", *errors);
    glEnable(GL_TEXTURE_SHADER_NV);
    glEndList();

    bind_specular_register_combiners = glGenLists(1);
    glNewList(bind_specular_register_combiners, GL_COMPILE);
    // Configure register combiners
    nvparse(
        "!!RC1.0                                           \n"
        "// Diffuse and specular lighting                  \n"
        "const0 = (.1, .1, .1, .1); // ambient intensity   \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // simple diffuse calculation N dot L     \n"
        "        spare0 = expand(tex3) . expand(col1);     \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // R = 2 * (N * (N dot L) - L/2)          \n"
        "        //   = 2 * N * (N dot L) - L              \n"
        "        discard = spare0 * expand(tex3);          \n"
        "        discard = half_bias(zero) * expand(col1); \n"
        "        spare1 = sum();                           \n"
        "        scale_by_two();                           \n"
        "    }                                             \n"
        "    alpha                                         \n"
        "    {                                             \n"
        "        // self_shadow = spare0.a = 8 * (N dot L) \n"
        "        discard = unsigned(spare0.b);             \n"
        "        discard = unsigned(spare0.b);             \n"
        "        spare0.a = sum();                         \n"
        "        scale_by_four();                          \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // spare0 = color * (diffuse + ambient)   \n"
        "        discard = col0 * unsigned(spare0);        \n"
        "        discard = col0 * const0;                  \n"
        "        spare0 = sum();                           \n"
        "    }                                             \n"
        "    alpha                                         \n"
        "    {                                             \n"
        "        // E = (0, 0, 1)                          \n"
        "        // R dot E = R dot (0, 0, 1) = R.b        \n"
        "        // spare1.a = (R dot E) ^ 2               \n"
        "        spare1.a = spare1.b * spare1.b;           \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // self_shadow * (R dot E) ^ 2            \n"
        "        spare1 = spare0.a * spare1.a;             \n"
        "    }                                             \n"
        "    alpha                                         \n"
        "    {                                             \n"
        "        // spare1.a = (R dot E) ^ 4               \n"
        "        spare1 = spare1 * spare1;                 \n"
        "    }                                             \n"
        "}                                                 \n"
        "// specular = (R dot E)^8                         \n"
        "final_product = spare1.a * spare1.a;              \n"
        "// out.rgb = self_shadow * (R dot E)^10 + diffuse \n"
        "out.rgb = spare1 * final_product + spare0.rgb;    \n"
        "out.a = tex3.a;                                   \n"
    );
    for (errors = nvparse_get_errors(); *errors; errors++)
        fprintf(stderr, "Register combiner error: %s\n", *errors);
    glEnable(GL_REGISTER_COMBINERS_NV);
    glEndList();

    bind_specular_with_normalization_register_combiners = glGenLists(1);
    glNewList(bind_specular_with_normalization_register_combiners, GL_COMPILE);
    // Configure register combiners
    nvparse(
        "!!RC1.0                                           \n"
        "// Diffuse and specular lighting                  \n"
        "// Light vector is renormalized per pixel         \n"
        "const0 = (.1, .1, .1, .1); // ambient intensity   \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // normalize L                            \n"
        "        spare0 = expand(col1) . expand(col1);     \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // continue normalizing L                 \n"
        "        discard = unsigned_invert(spare0) * half_bias(col1); \n"
        "        discard = expand(col1);                   \n"
        "        col1 = sum();                             \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // simple diffuse calculation N dot L     \n"
        "        spare0 = expand(tex3) . col1;             \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // R = 2 * (N * (N dot L) - L/2)          \n"
        "        //   = 2 * N * (N dot L) - L              \n"
        "        discard = spare0 * expand(tex3);          \n"
        "        discard = half_bias(zero) * col1;         \n"
        "        spare1 = sum();                           \n"
        "        scale_by_two();                           \n"
        "    }                                             \n"
        "    alpha                                         \n"
        "    {                                             \n"
        "        // self_shadow = spare0.a = 8 * (N dot L) \n"
        "        discard = unsigned(spare0.b);             \n"
        "        discard = unsigned(spare0.b);             \n"
        "        spare0.a = sum();                         \n"
        "        scale_by_four();                          \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // spare0 = color * (diffuse + ambient)   \n"
        "        discard = col0 * unsigned(spare0);        \n"
        "        discard = col0 * const0;                  \n"
        "        spare0 = sum();                           \n"
        "    }                                             \n"
        "    alpha                                         \n"
        "    {                                             \n"
        "        // E = (0, 0, 1)                          \n"
        "        // R dot E = R dot (0, 0, 1) = R.b        \n"
        "        // spare1.a = (R dot E) ^ 2               \n"
        "        spare1.a = spare1.b * spare1.b;           \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // self_shadow * (R dot E) ^ 2            \n"
        "        spare1 = spare0.a * spare1.a;             \n"
        "    }                                             \n"
        "    alpha                                         \n"
        "    {                                             \n"
        "        // spare1.a = (R dot E) ^ 4               \n"
        "        spare1 = spare1 * spare1;                 \n"
        "    }                                             \n"
        "}                                                 \n"
        "// specular = (R dot E)^8                         \n"
        "final_product = spare1.a * spare1.a;              \n"
        "// out.rgb = self_shadow * (R dot E)^10 + diffuse \n"
        "out.rgb = spare1 * final_product + spare0.rgb;    \n"
        "out.a = tex3.a;                                   \n"
    );
    for (errors = nvparse_get_errors(); *errors; errors++)
        fprintf(stderr, "Register combiner error: %s\n", *errors);
    glEnable(GL_REGISTER_COMBINERS_NV);
    glEndList();

    bind_diffuse_register_combiners = glGenLists(1);
    glNewList(bind_diffuse_register_combiners, GL_COMPILE);
    // Configure register combiners
    nvparse(
        "!!RC1.0                                           \n"
        "// Diffuse lighting only (no specular)            \n"
        "const0 = (.1, .1, .1, .1); // ambient intensity   \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // simple diffuse calculation N dot L     \n"
        "        spare0 = expand(tex3) . expand(col1);     \n"
        "        col1 = const0;                            \n"
        "    }                                             \n"
        "}                                                 \n"
        "// color_sum = clamp0to1(spare0 + col1)           \n"
        "//           = clamp0to1(N dot L + ambient)       \n"
        "clamp_color_sum();                                \n"
        "// out.rgb = ((N dot L) + ambient) * color        \n"
        "out.rgb = color_sum * col0;                       \n"
        "out.a = tex3.a;                                   \n"
    );
    for (errors = nvparse_get_errors(); *errors; errors++)
        fprintf(stderr, "Register combiner error: %s\n", *errors);
    glEnable(GL_REGISTER_COMBINERS_NV);
    glEndList();


    bind_diffuse_with_normalization_register_combiners = glGenLists(1);
    glNewList(bind_diffuse_with_normalization_register_combiners, GL_COMPILE);
    // Configure register combiners
    nvparse(
        "!!RC1.0                                           \n"
        "// Diffuse lighting only (no specular)            \n"
        "// Light vector is renormalized per pixel         \n"
        "const0 = (.1, .1, .1, .1); // ambient intensity   \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // normalize L                            \n"
        "        spare0 = expand(col1) . expand(col1);     \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // continue normalizing L                 \n"
        "        discard = unsigned_invert(spare0) * half_bias(col1); \n"
        "        discard = expand(col1);                   \n"
        "        col1 = sum();                             \n"
        "    }                                             \n"
        "}                                                 \n"
        "{                                                 \n"
        "    rgb                                           \n"
        "    {                                             \n"
        "        // simple diffuse calculation N dot L     \n"
        "        spare0 = expand(tex3) . col1;     \n"
        "        col1 = const0;                            \n"
        "    }                                             \n"
        "}                                                 \n"
        "// color_sum = clamp0to1(spare0 + col1)           \n"
        "//           = clamp0to1(N dot L + ambient)       \n"
        "clamp_color_sum();                                \n"
        "// out.rgb = ((N dot L) + ambient) * color        \n"
        "out.rgb = color_sum * col0;                       \n"
        "out.a = tex3.a;                                   \n"
    );
    for (errors = nvparse_get_errors(); *errors; errors++)
        fprintf(stderr, "Register combiner error: %s\n", *errors);
    glEnable(GL_REGISTER_COMBINERS_NV);
    glEndList();
}

static void menu( int v )
{
    b[v] = !b[v];

    switch (v)
    {
    case 27:
    case 'q':
        exit(0);
    case 'd':
    case 'D':
        if ('d' == v)
            dim++;
        else if (dim > 1)
            dim--;
        glDeleteLists(draw_solid_sphere, 1);
        create_sphere_dl();
        break;
    case 'l':
        light_radius += light_radius_step;
        break;
    case 'L':
        if (light_radius > min_light_radius)
            light_radius -= light_radius_step;
        break;
    case ' ':
        if(b[v])
            glutIdleFunc(idle);
        else
            glutIdleFunc(0);
        break;
    }


    glutPostRedisplay();
}

static void key(unsigned char k, int x, int y)
{
    menu((int)k);
}

static void idle()
{
    theta += theta_step;
    glutPostRedisplay();
}

// The next two functions define a (probably excessively complicated)
// color mapping scheme for 2D coordinates.  They are collectively
// used to select interesting colors for the sprites.
static void calc_rgb_color(GLfloat u, vec3f& color)
{
    /* These should look familiar... (Second degree Bernstein basis) */
    color[0] = (1.0 - u) * (1.0 - u);
    color[1] = 2.0 * u * (1.0 - u);
    color[2] = u * u;
}

static void set_rgb_color(vec3f& color, int i, int j, int dim)
{
    vec3f color_x, color_y;

    calc_rgb_color((float)i/(float)dim, color_x);
    calc_rgb_color((float)j/(float)dim, color_y);
    color[0] = clamp_0_to_1(.7 * (color_x[0] + color_y[1]) * 3.0 / 2.0);
    color[1] = clamp_0_to_1(.7 * (color_x[1] + color_y[0]) * 3.0 / 2.0);
    color[2] = clamp_0_to_1(.7 * (color_x[2] + color_y[2]) * 3.0 / 2.0);
}






// p mu normalisiert sein
void GetOtherBaseVec( vec3f p, vec3f &a, vec3f &b )
{
	if(p[2]<-0.5f || p[2]>0.5f)
	{
		a[0]=p[2];
		a[1]=p[1];
		a[2]=-p[0];
//		a.normalize();	// effizienter machbar
		b=p.cross(a);
	}
	else
	{
		a[0]=p[1];
		a[1]=-p[0];
		a[2]=p[2];
//		a.normalize();	// effizienter machbar
		b=p.cross(a);
	}
}



static vec3f g_position[10000];
static float g_radius[10000];


static void display()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glPushMatrix();
    camera.apply_inverse_transform();

    object.apply_transform();

    glCallList(bind_vertex_program);

    if (b['t'])
        glCallList(bind_texture_shaders_depth);
    else
        glCallList(bind_texture_shaders_no_depth);

    if (b['s'])
    {
        if (b['n'])
            glCallList(bind_specular_with_normalization_register_combiners);
        else
            glCallList(bind_specular_register_combiners);
        float specular[] = { 1, 1, 1, 1};
        glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
    }
    else
    {
        if (b['n'])
            glCallList(bind_diffuse_with_normalization_register_combiners);
        else
            glCallList(bind_diffuse_register_combiners);
        float specular[] = { 0, 0, 0, 1};
        glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
    }

    glAlphaFunc(GL_GEQUAL, .9);
    if (b['a'])
        glEnable(GL_ALPHA_TEST);
    else
        glDisable(GL_ALPHA_TEST);

    if (b['z'])
        glEnable(GL_DEPTH_TEST);
    else
        glDisable(GL_DEPTH_TEST);

    if (b['w'])
    {
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glDisable(GL_ALPHA_TEST);
    }
    else
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    // The vertex program is responsible for calculating the vector from the sprite
    // corners to the light, normalizing it, range compressing it and putting it into
    // secondary color.
    vec3f light_position;
    float phi = theta / (2. * PI) * theta_step;
    light_position[0] = light_radius * sin(theta);
    light_position[1] = light_radius * cos(theta) * cos(phi);
    light_position[2] = light_radius * cos(theta) * sin(phi);

    float offset = -1. + 1. / (float)dim;
    float delta = 2. / (float)dim;

    if (b['g'])
    {
        // Draw geometric spheres with per-vertex lighting
        float light[4];
        light[0] = light_position[0];
        light[1] = light_position[1];
        light[2] = light_position[2];
        light[3] = 1.0;
        glLightfv(GL_LIGHT0, GL_POSITION, light);
        float ambient[] = { .1, .1, .1, 1.};
        glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
        glMaterialf(GL_FRONT, GL_SHININESS, 32.);
        float global_ambient[] = { 0, 0, 0, 1};
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
        glDisable(GL_VERTEX_PROGRAM_NV);
        glDisable(GL_TEXTURE_SHADER_NV);
        glDisable(GL_REGISTER_COMBINERS_NV);
        glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);
        glDisable(GL_ALPHA_TEST);

        for (int j = 0; j < dim; j++)
        {
            for (int i = 0; i < dim; i++)
            {
                vec3f diffuse_color;
                set_rgb_color(diffuse_color, i, j, dim);
                glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, &diffuse_color[0]);
                glTranslatef(offset + (float)i * delta, offset + (float)j * delta, 0);
                glCallList(draw_solid_sphere);
                glTranslatef(-(offset + (float)i * delta), -(offset + (float)j * delta), 0);
            }
        }

        // Draw the light position
        // Define phantom lighting at (0, 0, 0) to illuminate the light
        float phantom_light_position[] = {0, 0, 0, 1};
        glLightfv(GL_LIGHT0, GL_POSITION, phantom_light_position);
        // Make the light plain white in color
        float material[] = {1, 1, 1, 1};
        glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
        glTranslatef(light_position[0], light_position[1], light_position[2]);
        // Draw the light as a sphere, and give it a hardcoded radius of .1
        glutSolidSphere(.1, 100, 100);
        glTranslatef(-light_position[0], -light_position[1], -light_position[2]);
    }
    else
    {
        // Draw depth sprites with per-pixel lighting
        glDisable(GL_LIGHTING);
        glSecondaryColor3fEXT(light_position[0], light_position[1], light_position[2]);
        glBegin(GL_QUADS);
/*
		// Draw the sprites in a dim X dim square
        for (int j = 0; j < dim; j++)
        {
            for (int i = 0; i < dim; i++)
            {
                vec3f diffuse_color;
                set_rgb_color(diffuse_color, i, j, dim);
                glColor3f(diffuse_color[0], diffuse_color[1], diffuse_color[2]);
                glTexCoord4f(offset + (float)i * delta, offset + (float)j * delta, 0, sphere_radius);
                // Draw sprite corners
                glVertex4f(-1, -1, 0, 0);
                glVertex4f(-1,  1, 0, 0);
                glVertex4f( 1,  1, 0, 0);
                glVertex4f( 1, -1, 0, 0);
            }
        }
*/
   
		// wormring
		{
			for(float f=0.0f;f<50.0f;f+=0.25f*3.0f*0.5f*3.0f)
			{
				vec3f mid=vec3f(sin(f*0.25f)*0.4f,cos(f*0.25f)*0.3f+0.08f,cos(f*0.25f)*0.15f+1.0f);
				float r=(sin(f*0.6f)+1.1f)*0.08f;

				float radius=r;
				vec3f pos=mid;

				pos=pos*2.0+vec3f(0,0,-2);radius*=2.0;


                glColor3f(0.8f, 0.1f, 0.0f);
                glTexCoord4f(pos[0], pos[1], pos[2], radius);

				// Draw sprite corners
                glVertex4f(-1, -1, 0, 0);
                glVertex4f(-1,  1, 0, 0);
                glVertex4f( 1,  1, 0, 0);
                glVertex4f( 1, -1, 0, 0);
			}
		}		


		static int Count=0;


		if(!Count)
		{
			// umschlingende Ringe (unregelmig)
			for(float f=0.0f;f<50.0f;f+=0.25f*0.022f*0.4f)
//		for(float f=0.0f;f<50.0f;f+=0.25f*0.022f)
	//		for(float f=0.0f;f<50.0f;f+=0.25f*0.22f)
			{
				float k=f+sin(f)*4.3f;

				vec3f mid=vec3f(sin(k*0.25f)*0.4f,cos(k*0.25f)*0.3f+0.08f,cos(k*0.25f)*0.15f+1.0f);
				vec3f mid2=vec3f(sin(k*0.25f+0.1f)*0.4f,cos(k*0.25f+0.1f)*0.3f+0.08f,cos(k*0.25f+0.1f)*0.15f+1.0f);
				float r=(sin(k*0.6f)+1.1f)*0.08f;

				vec3f direction=mid-mid2;		direction.normalize();
				vec3f v1,v2;

				GetOtherBaseVec(direction,v1,v2);

				float s=sin(f*12.0f*2.0f)+cos(f*46.0f)*0.3f;
				float c=cos(f*12.0f*2.0f)+sin(f*22.0f)*0.3f;


				vec3f pos=mid+v1*r*s+v2*r*c;
				float radius=0.015f+(sin(f*42.0f)+sin(f*330.0f)*0.4f+0.4)*0.008f;


				pos=pos*2.0+vec3f(0,0,-2);radius*=2.0;



				bool drawit=true;
				{
				// Wurmring
					for(float f=0.0f;f<50.0f;f+=0.25f*3.0f*0.5f*3.0f)
					{
						vec3f mid=vec3f(sin(f*0.25f)*0.4f,cos(f*0.25f)*0.3f+0.08f,cos(f*0.25f)*0.15f+1.0f);
						float r=(sin(f*0.6f)+1.1f)*0.08f;
						float radius2=r;
						vec3f pos2=mid;

						pos2=pos2*2.0+vec3f(0,0,-2);radius2*=2.0;

						if(radius2>radius)
						{
							float dist=(pos2-pos).length();

							if(dist+radius<radius2)
								{ drawit=false;break;}
						}
					}	
				}


				if(drawit)
				{
					g_position[Count]=pos;
					g_radius[Count]=radius;
					Count++;
				}
			}

		}
		else
		{
			srand(10032);
			for(int i=0;i<Count;i++)
			{
				float gg=(rand()%1024)*0.001f;

				glColor3f(1.0f, gg, gg);

				glTexCoord4f(g_position[i][0], g_position[i][1], g_position[i][2], g_radius[i]);

				// Draw sprite corners
				glVertex4f(-1, -1, 0, 0);
				glVertex4f(-1,  1, 0, 0);
				glVertex4f( 1,  1, 0, 0);
				glVertex4f( 1, -1, 0, 0);
			}
		}


		// Draw the light position
		// Define phantom lighting at (0, 0, 0) to illuminate the light
		glSecondaryColor3fEXT(0, 0, 0);
		// Make the light plain white in color
		glColor3f(1., 1., 1.);
		// Draw the light as a sprite, and give it a hardcoded radius of .1
		glTexCoord4f(light_position[0], light_position[1], light_position[2], .1);
		// Draw sprite corners
		glVertex4f(-1, -1, 0, 0);
		glVertex4f(-1,  1, 0, 0);
		glVertex4f( 1,  1, 0, 0);
		glVertex4f( 1, -1, 0, 0);
		glEnd();
	}

    glPopMatrix();

    glutSwapBuffers();

#define MAX_FRAMES  10 //30
    framecount++;

//    if(b['r'])
    {
        double cum_fps, fps;

        if(framecount >= MAX_FRAMES)
        {
            QueryPerformanceCounter(&newCount);

            fps = (double)frequency.QuadPart
                            / (double)(newCount.QuadPart - oldCount.QuadPart);
            cum_fps = fps * MAX_FRAMES;

		    static char buf[48];

		    sprintf(buf,"fps: %.3f\n", (float)cum_fps);

			printf(buf);

            framecount = 0;
            cum_fps = 0;
            QueryPerformanceCounter(&oldCount);
        }
    }
}
